/*
2LMS2Rabbit
Copyright 2006 Stephan Poehlsen, Hannes Bistry

    This file is part of 2LMS2Rabbit.

    2LMS2Rabbit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    2LMS2Rabbit is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with 2LMS2Rabbit.  If not, see <http://www.gnu.org/licenses/>.

*/



// ex: set ts=3:  -*- tab-width:3 -*-

//debug messages...
#define DEBUG 1


//if address (2nd byte of telegram) of LMS at port F is replaced by 0x81 and CRC is patched?
//our LMS refused to changed his address so we "mark" the telegrams to differentiate the 2 LMS
#define SWITCH_PORTF_ADDRESS_AND_CRC 1



// ******************************************************************
// ******************************************************************
// ******************************************************************

//our library that contains a list of patch value to correct CRC if we modify 2nd byte
#use "crc_offset.lib"


// ******************************************************************

#ifdef DEBUG
#define report(port,x)	udpdebug(port,x); printf("%c: %s\n", port, x)
void udpdebug(char port, char *data);
#else
#define report(port,x)
#endif

/*****************************
 *** UDP                   ***
 *****************************/

/*no default config*/
#define TCPCONFIG		0

/*one Ethernet Device */
#define USE_ETHERNET		1

/* configuration of Ethernet devices */
/*#define IFCONFIG_ETH0 \
	IFS_IPADDR, aton("134.100.13.229"), \
 	IFS_NETMASK, aton("255.255.255.0"), \
 	IFS_ROUTER_SET, aton("134.100.13.250"), \
	IFS_UP
*/
//replace it with your address/subnet
#define IFCONFIG_ETH0 \
	IFS_IPADDR, aton("192.168.0.2"), \
 	IFS_NETMASK, aton("255.255.255.0"), \
 	IFS_ROUTER_SET, aton("192.168.0.1"), \
	IFS_UP

/* Define the number of socket buffers that will be allocated for
 * UDP sockets. */
#define MAX_UDP_SOCKET_BUFFERS 2

/* what local UDP port to use - packets we send come from here */
#define LOCAL_PORT	2222

/* If it is set to all '255's, it will be a broadcast packet. */
//if you do not want to use multicast, exchange it with an unicast address
//i.e. 192.168.0.xxx
#define REMOTE_IP			"224.0.0.23"
//#define REMOTE_IP			"255.255.255.255" /*broadcast*/

/* the destination port to send to */
#define REMOTE_PORT	2222

#define DISABLE_DNS 1  //no DNS - static IP
#define DISABLE_TCP 1  //no TCP-stack - would cause overhead...

/* default 600*/
#define ETH_MTU 1500

#memmap xmem
#use "dcrtcp.lib"

udp_Socket sock_out;
udp_Socket sock_in;

/************************************
 * RS422
 **************************************/

#define FAST_INTERRUPT 0
#define BUFFER_COUNT 8	//number of buffers in ringbuffer
#define BUFFER_MASK 0x07   //BUFFER_COUNT-8 - is ANDed to 

//used for minimal distance check
int warndist;
int alarmdist;


//MAIN struct of program, one for each LMS
typedef struct {
	char port;//e or f - we use these serial ports because the have a FIFO
   unsigned char buf[768*(1+BUFFER_COUNT)];//memory for some(1+BUFFER_COUNT) telegrams
	//we use 1 more buffer to exchange buffers-with this architecture we can do asynchronous preprocessing-not used now

	char **pointer;	// array with pointers to each telegram(1+BUFFER_COUNT)
	char *work;			// pointer to buffer-element not used/excluded for async processing
								
	char *in;				// pointer to next element to be written
								// if porte_in == porte_out ==> Buffer full
	char *out;			// pointer to element that HAS been read already
								//porte_out+1 == porte_in ==> Puffer leer
								// inc modulo BUFFER_COUNT, and if !=porte_in (new data) send UDP paket
	
	char extract;        //if next element should be exchanged for preprocessing - not needed
	//buffer for telegrams TO LMS
	char outbuf[256];    //Telegrammpuffer, fuer Telegramme an den LMS
   char outbuf_in;      //Offset fuer das Schreiben in den outbuf
   char outbuf_out;     //Offset fuer das Lesen aus dem outbuf
	
	
   int maxfs;         //variable to detect overruns of main buffer  
   char run;          //state variable if LMS should be set to their continuous mode 
   int lms_state;    //state of automat      
	
   int time;              //Timestamp of last action in automat
   
   //for preprocessing
   int portwarn[361];     //array of minimal distance for each angle - alarm level 1
   int portalarm[361];    //array of minimal distance for each angle - alarm level 2
   char distcheck;        //actual warnlevel
   int posx;              //monting position of LMS to centre of robot X
   int posy;              //monting position of LMS to centre of robot Y
   int angle;             //angle to foreward driving direction
   //!for preprocessing
} fsm;

fsm porte, portf;//instantiate 2 of these


//define some variables to access them easier inside ISR(no overhead like inside a struct)
char *porte_pointer[1+BUFFER_COUNT];	// Array with X buffers to telegrams
char *porte_position;	// actual writing position within telegram  
int isre_state;		//state of ISR automat
int porte_length;		//telegram length
char porte_in;		//0 till BUFFER_COUNT-1
char porte_out;		//0 till BUFFER_COUNT-1

unsigned porte_crc;	// for CRC-check inside ISR
char porte_crc_olddata;


//same for port F

unsigned portf_crc;	
char portf_crc_olddata;

char *portf_pointer[1+BUFFER_COUNT];	// Array mit 5 ZEIGERN auf Zeichenketten (Telegramme)
char *portf_position;
int isrf_state;
int portf_length;
char portf_in;
char portf_out;


int porte_bogus, portf_bogus;
int porte_over, portf_over;

unsigned time, led;

//different LMS-automat states
enum {
	LMS_IDLE,
   LMS_9K6pre,
   LMS_9K6prewait,
   LMS_9K6,
   LMS_9K6wait,
   LMS_19K2pre,
   LMS_19K2prewait,
   LMS_19K2,
   LMS_19K2wait,
   LMS_38K4pre,
   LMS_38K4prewait,
   LMS_38K4,
   LMS_38K4wait,
   LMS_500Kpre,
   LMS_500Kprewait,
   LMS_500K,
   LMS_500Kwait,
   LMS_FOUND,
   LMS_FOUNDwait,
   LMS_SYNC,
   LMS_SYNCwait,
   LMS_CHECK,
   LMS_CHECKwait,
   LMS_WORK,
   LMS_WORKwait,
   LMS_DATA,
   LMS_DATAwait,
   LMS_KILL,
   LMS_KILLwait,
   LMS_KILLpre,
   LMS_KILLprewait
};

int port_send_packet(fsm *port);
void porte_set_speed(char s);
void portf_set_speed(char s);
void port_send_telegram(fsm *port, char* data, char len);
void update_distarray(fsm* port);

#define STWAIT 60
#define WCYCLES 10   // times that we wait for STWAIT ms until timeout


#use "isr_asm.lib"
#use "lms_automat.lib"


//for preprocessing we pre-calculate a tabular - not needed
float sintab[541]; // halbe grad schritte
const float *const sint = sintab;
const float *const cost = sintab+180;

/***************************************************************************
 **************************************************************************/


int send_udp(void *addr, int len) {//send routine
	if (udp_send(&sock_out, addr, len) < 0) {
		printf("Error sending datagram!  Closing and reopening socket...\n");
		sock_close(&sock_out);
		if(!udp_open(&sock_out, LOCAL_PORT, resolve(REMOTE_IP), REMOTE_PORT, NULL)) {
			printf("udp_open failed!\n");
		}
   }
	tcp_tick(NULL);
	return 1;
}

void port_send_telegram(fsm *port, char* data, char len) {//send telegrams from PC to LMS 
   char offset, i;
   offset = port->outbuf_in;
   for (i = 0; i < len; i++) {
		port->outbuf[offset] = data[i];
   	offset++;
   }
   port->outbuf_in = offset;
	send_udp(data, len);
}

int port_send_packet(fsm *port) {//process telegrams
	int retval, length;
#if SWITCH_PORTF_ADDRESS_AND_CRC
   int crc;
#endif
   char out, buf[361*4+2];
   char *temp;
	int i, i2, cnt, dist, dist2, mint;
	unsigned den, densum, dencnt, denold, denangle, denoffset;
   char *bufin, alarm_value;
   unsigned bufinrest;
   int dendistsum, dendist;
   // char timelow, timelowold, timehigh, timehighold;

	out = (*(port->out)+1)&BUFFER_MASK;
   if (out == *(port->in)) return 1;

   length = port->pointer[(int)out][2] // LSB
   			+ (port->pointer[(int)out][3]<<8) // MSB
            + 6;

   if (port->port == 'f') {
		port->pointer[out][1] |= 0x01;
#if SWITCH_PORTF_ADDRESS_AND_CRC
		if (length > 701 && length < 740) crc = crc_offset700[length-702];
		else if (length < 162) crc = crc_offset[length-2];
      port->pointer[out][length-2] ^= (crc & 0xff);
      port->pointer[out][length-1] ^= crc>>8;
#endif
   }

	// send the packet
	retval = send_udp(port->pointer[(int)out], length);
   
//FROM here on preprocessing stuff
   
   /*
	// wert der RealTimeClock abfragen
   WrPortI(0x02, NULL,0x00);
   timelowold=RdPortI(0x02);
   timehighold=RdPortI(0x03);
   */

   // Scanwerte checken checken
   if (length == 732) {
      buf[0] = 0x02;
      buf[1] = port->pointer[out][1]+0x10;
      buf[4] = 0xB3;

	   bufin = &buf[5];

	   alarm_value = 0;
		cnt = 0;
      densum = 0;
      dencnt = 0;
      denold = 0;
      denangle = 0;
      dendist = 0;
      dendistsum = 0;
      temp = port->pointer[(int)out]+7;
	   for (i = 0; i < 361; i++) {
      	dist = *(temp++);
			dist |= ((*temp)&0x1f)<<8;
      	den = (unsigned) ((*(temp++)) >> 5);

         // uberwachen des Distance
			if (port->distcheck) {
				if (!alarm_value && (dist < port->portwarn[i])) {
              	alarm_value = 1;
            	if (dist < port->portalarm[i]) alarm_value = 2;
				} else if (alarm_value == 1) {
            	if (dist < port->portalarm[i]) alarm_value = 2;
            }
         }

         if (denold &&
             (!den ||(dendist-100 > dist) || (dendist+100 < dist)))
             {
            // Marke berechnen und sichern
          	cnt++;
				dendist = (dendist+(dendistsum/(int)densum));
				*(bufin++) = dendist & 0xff;
            *(bufin++) = dendist >> 8;
				denangle = denoffset*60+(denangle*60/densum);
				*(bufin++) = denangle & 0xff;
            *(bufin++) = denangle >> 8;

				//printf("%3u: winkel:%u dist:%d dendistsum=%d dencnt=%u\n", i,
            // 		 angle, dendist, dendistsum, dencnt);
            dendistsum = 0;
				densum = 0;
            dencnt = 0;
            denangle = 0;
            dendist = 0;
         }

         if (den) {
            if (!dendist) {
            	denoffset = i;
	            dendist = dist;
            }
            dendistsum += (dist-dendist)*den;
         	densum += den;
            dencnt++;
            denangle += (i-denoffset)*den;
         }
         denold = den;

		}

      if (denold) {
         // Marke berechnen und sichern
        	cnt++;
			dendist = (dendist+(dendistsum/(int)densum));
			*(bufin++) = dendist & 0xff;
         *(bufin++) = dendist >> 8;
			denangle = denoffset*60+(denangle*60/densum);
			*(bufin++) = denangle & 0xff;
         *(bufin++) = denangle >> 8;
      }

      // LEDs setzen
      if (port->port == 'e') {
        led = (led & 0xf8) | (1<<alarm_value);
      } else {
        led = (led & 0xc7) | (0x20>>alarm_value);
      }
      WrPortI(PADR, NULL, led);

      // Marken per udp senden
		*(bufin++) = alarm_value;
      cnt = (cnt<<2)+2;
      buf[2] = (cnt)&0xff;
      buf[3] = (cnt>>8)&0xff;
		send_udp(buf, (cnt+4));
   } // ende Messdaten verarbeiten

   /*
   // Wert der RealTimeCloeck abfrage
   WrPortI(0x02, NULL,0x00);
   timelow=RdPortI(0x02);
   timehigh=RdPortI(0x03);

   // Zeitdifferenz ausgeben
	snprintf(buf, sizeof(buf), "Zeitmessung: %d",
            ((timehigh-timehighold)<<8)+timelow-timelowold);
   send_udp(buf, strlen(buf));
   */

   // packet vielleicht extrahieren
   if (port->extract) {
   	temp = port->pointer[out];
      port->pointer[out] = port->work;
      port->work = temp;
      port->extract = 0;
   }
   //untill herer was the preprocessing
   
   
   /* porte_out weiter setzen */
   *(port->out) = out;


	return retval;
}


/***************************************************************************
 **************************************************************************/

#if DEBUG
void udpdebug(char port, char *data) {//debug messages over UDP
	char buf[256];
   buf[0] = 0x02;
   buf[1] = 0xff;
   buf[3] = 0x00;
   snprintf(&buf[4], 250, "%c: %s", port, data);
   buf[2] = strlen(&buf[4])-2;
	send_udp(buf, buf[2]+6);
}
#endif

/***************************************************************************
 **************************************************************************/


void porte_set_speed(char s) {
	switch (s) {
   case 5: // 500K
		WrPortI(TAT2R, NULL, 12); // PortE:13*8*500KBaud
 		BitWrPortI(TACR, &TACRShadow, 0, 2);	// TimerA2: mainClk
		report('e', "PortE set to 500K Baud");
	   break;
   case 1: // 19K2
		WrPortI(TAT2R, NULL, 1); // PortE:168*8*2
		BitWrPortI(TACR, &TACRShadow, 1, 2);	// TimerA2: TimerA1
		report('e', "PortE set to 19K2 Baud");
	   break;
   case 9: // 9K6
		WrPortI(TAT2R, NULL, 3); // PortE:168*8*4
		BitWrPortI(TACR, &TACRShadow, 1, 2);	// TimerA2: TimerA1
		report('e', "PortE set to 9K6 Baud");
	   break;
   case 3: // 38K4
		WrPortI(TAT2R, NULL, 0); // PortE:168*8*1
		BitWrPortI(TACR, &TACRShadow, 1, 2);	// TimerA2: TimerA1
		report('e', "PortE set to 38K4 Baud");
	   break;
   case 0:
		WrPortI(TAT2R, NULL, 0xff);
		BitWrPortI(TACR, &TACRShadow, 1, 2);	// TimerA2: TimerA1
		report('e', "PortE set to 150 Baud");
	   break;
   }
   porteflush();
}

void portf_set_speed(char s) {
	switch (s) {
   case 5: // 500K
		WrPortI(TAT3R, NULL, 12); // PortF:13*8*500KBaud
 		BitWrPortI(TACR, &TACRShadow, 0, 3);	// TimerA3: mainClk
		report('f', "PortF set to 500K Baud");
	   break;
   case 1: // 19K2
		WrPortI(TAT3R, NULL, 1); // PortF:168*8*2
		BitWrPortI(TACR, &TACRShadow, 1, 3);	// TimerA3: TimerA1
		report('f', "PortF set to 19K2 Baud");
	   break;
   case 9: // 9K6
		WrPortI(TAT3R, NULL, 3); // PortF:168*8*4
		BitWrPortI(TACR, &TACRShadow, 1, 3);	// TimerA3: TimerA1
		report('f', "PortF set to 9K6 Baud");
	   break;
   case 3: // 38K4
		WrPortI(TAT3R, NULL, 0); // PortF:168*8*1
		BitWrPortI(TACR, &TACRShadow, 1, 3);	// TimerA3: TimerA1
		report('f', "PortF set to 38K4 Baud");
	   break;
   case 0: // 150 BAUD
		WrPortI(TAT3R, NULL, 0xff); // PortF:168*8*1
		BitWrPortI(TACR, &TACRShadow, 1, 3);	// TimerA3: TimerA1
		report('f', "PortF set to 150 Baud");
	   break;
   }
   portfflush();
}

void fill_info(char *buf, int len) {//generate some status information about buffer fill level
	snprintf(
   		buf, len,
         "E:: run:%c btrash:%u ttrash:%u fill:%u  F:: run:%c btrash:%u ttrash:%u fill: %u",
    		(porte.lms_state == LMS_DATAwait || porte.lms_state == LMS_DATA)+'0',
         porte_bogus, porte_over, porte.maxfs,
     		(portf.lms_state == LMS_DATAwait || portf.lms_state == LMS_DATA)+'0',
         portf_bogus, portf_over, portf.maxfs);
}
// UDP telegram receive
int receive_packet(void) {
	static char buf[128];
   int len;
	#GLOBAL_INIT
   { memset(buf, 0, sizeof(buf)); }

	if (-1 == (len = udp_recv(&sock_in, buf, sizeof(buf)))) {
		//report('s', "len == -1"); send_udp("len == -1", 9);
		return 0; // no packet read. return
	}
   //if (len < 0) {	report('s', "len < -1"); send_udp("len < 0", 7); return 0; }
   //printf("Received-> %s\n",buf);

   switch (buf[0]){
   	case 'r':
      	report('s', "now resetting target...");
         if (buf[1]== 'e' && buf[2]=='s' && buf[3]=='e' && buf[4]=='t')forceSoftReset();
         break;
      case 'i':
      	report('s', "status info...");
         fill_info(buf, sizeof(buf));
         send_udp(buf, strlen(buf));
      	break;
      case 'd': // Telegramm will be send to LMS (after e or f)
     		if (buf[1]=='e')  port_send_telegram(&porte, &buf[2],(len-2));
     		if (buf[1]=='f')  port_send_telegram(&portf, &buf[2],(len-2));
         break;
      case 'k': // kill continoues data
         if (buf[1]=='e') porte.run = 0;//telegram "ke" stops LMS at e
         if (buf[1]=='f') portf.run = 0;
         break;
      case 's': // start continoues data
         if (buf[1]=='e') porte.run = 1;
         if (buf[1]=='f') portf.run = 1;
         break;
      case 'p': // change scanner position data (porte.x; y;angle;portf.x ,y,angle) - only relevant for preprocessing
			if (len == 13) {
	         porte.posx = buf[1]|(buf[2]<<8);
	         porte.posy = buf[3]|(buf[4]<<8);
	         porte.angle = buf[5]|(buf[6]<<8);
	         portf.posx = buf[7]|(buf[8]<<8);
	         portf.posy = buf[9]|(buf[10]<<8);
	         portf.angle = buf[11]|(buf[12]<<8);
	         update_distarray(&porte);
				update_distarray(&portf);
         }
         break;
      case 'w'://change warndistance values
			if (len == 5) {
	         warndist = buf[1]|(buf[2]<<8);
	         alarmdist = buf[3]|(buf[4]<<8);
	         update_distarray(&porte);
				update_distarray(&portf);
         }
         break;
      default:
      report('s', "unknown command...");
   }

	memset(buf, 0, sizeof(buf));
	return 1;
}


#use "async_preproc.lib"

//this is only relevant if you want to do async processing of a scan
int work_state;
cofunc int work_scan() {
//   report('s', "work in progress");
	switch (work_state) {
   case -1: // warte auf IDLE in den send_port_packet()
      //report('s', "state -1");
		if (!porte.extract && !portf.extract)
      	work_state = 1;
      break;
   case 1: // fordere Daten an
   	//report('s', "state 1");
		porte.extract = 1;
		portf.extract = 1;
      work_state = 2;
      break;
   case 2: // warte auf Daten
   	//report('s', "state 2");
		if (!porte.extract && !portf.extract)
      	work_state = 3;
      break;
	case 3:
   	//report('s', "state 3");
   	if (((porte.work[2] | porte.work[3]<<8) == 726) &&
          ((portf.work[2] | portf.work[3]<<8) == 726)) {
			work_state = 4;
         break;
      }
      report('s', "wrong telegram in work");
      work_state = -1;
   	break;
	case 4:
   	//report('s', "state 4");
   	//minimaler_abstand();
      ap_median(3);
      work_state = -1;
      break;
   }


	return 1;
}
//until here

//calculation of warning distances
void update_distarray(fsm* port) {
	float p, q, alpha;
   unsigned l;
   for (l = 0; l < sizeof(port->portwarn)/sizeof(int); l++) {
		alpha = ((float)port->angle+l/2.0)*PI/180.0;
      p = (float)port->posx * cos(alpha) + (float) port->posy * sin(alpha);
      q = (float)port->posx*(float)port->posx + (float)port->posy*(float)port->posy;
		port->portwarn[l] = (int) (-p + sqrt(p*p - (q-(float)warndist*(float)warndist)));
		port->portalarm[l] = (int) (-p + sqrt(p*p - (q-(float)alarmdist*(float)alarmdist)));
		//printf("%u: w=%d a=%d\n",l,port->portwarn[l],port->portalarm[l]);
   }
}
//!calculation of warning distances

#if DEBUG
char debug_buf[200];
#endif

//useless stability test under heavy load
cofunc void cpuburn() {
	int i, i2;
   float h;
   h = 13.34;
   for(;;) {
   	for (i2=0;i2<500;i2++) {
	   	for (i=0;i<20;i++) {
   	   	h=h*(h+11.23);
         }
	      yield;
      }
      printf("########### burning ##############\n");
   }
}
//!useless

void main() {
	int l, s;

#GLOBAL_INIT
{
   led = 0;
   work_state = -1;
	time = 0;
   porte.lms_state = LMS_IDLE;
   portf.lms_state = LMS_IDLE;

	porte.run = 1;//always try to start lms
	portf.run = 1;

	porte.maxfs = 0;
	portf.maxfs = 0;

   porte.port = 'e';
   portf.port = 'f';

   porte.posx = 0;
   portf.posx = 0;

   porte.posy = 336;
   portf.posy = -336;

   porte.angle = 0;
   portf.angle = 180;

   warndist = 1000;
   alarmdist = 600;

   porte.distcheck = 1;
   portf.distcheck = 1;

   porte.extract = 0;
   portf.extract = 0;

	porte.outbuf_in = 0;
	porte.outbuf_out = 0xff;
	portf.outbuf_in = 0;
	portf.outbuf_out = 0xff;

   porte.in = &porte_in;
   porte.out = &porte_out;
   portf.in = &portf_in;
   portf.out = &portf_out;
}

	update_distarray(&porte);
	update_distarray(&portf);
#if SWITCH_PORTF_ADDRESS_AND_CRC
	printf("SWITCH_PORTF_ADDRESS_AND_CRC\n");
#endif

#if __SEPARATE_INST_DATA__
	printf("__SEPARATE_INST_DATA__\n");
#endif

	printf("Running ...\n");

   /////////////////////////////////////
   // UDP
	printf("Init socket ...");
	sock_init();
   printf(" done\n");

	printf("Opening UDP socket out...");
	if(!udp_open(&sock_out, LOCAL_PORT, resolve(REMOTE_IP), REMOTE_PORT, NULL)) {
		printf("udp_open failed!\n");
		exit(0);
	}
   printf(" done\n");

	printf("Opening UDP socket in...");
   if(!udp_open(&sock_in, LOCAL_PORT, 0, 0, NULL)) {
		printf("udp_open failed!\n");
		exit(0);
	}

   printf(" done\n");

   ////////////////////////////////////////
   // RS422
	// initialiseren von PortE Daten
	for (l = 0; l <= BUFFER_COUNT; l++) {
		porte_pointer[l] = porte.buf+(l*768);
		portf_pointer[l] = portf.buf+(l*768);
	}
	porte.work = porte_pointer[BUFFER_COUNT];
	porte_in = 0;
	porte_out = BUFFER_COUNT-1;
	portf.work = portf_pointer[BUFFER_COUNT];
	portf_in = 0;
	portf_out = BUFFER_COUNT-1;

   porte.pointer = porte_pointer;
   portf.pointer = portf_pointer;




#if __SEPARATE_INST_DATA__ && FAST_INTERRUPT
	interrupt_vector sere_intvec porte_isr;
	interrupt_vector serf_intvec portf_isr;
#else
	SetVectIntern(0x1C, porte_isr);	// set up ISR
	SetVectIntern(0x1D, portf_isr);	// set up ISR
#endif
  	report('s', "ISR registered ...");


	port_init();	// ASM-Code to initialize automat for ISR

   WrPortI(SPCR, NULL, 0x84); // PortA as outputport 
   WrPortI(PADR, NULL, led);

   printf("3\n");
	WrPortI(TAT9R, NULL, 1);
	printf("7\n");
   WrPortI(PWM0R, NULL, 0);
   printf("1\n");
   WrPortI(PWL0R, NULL, 0);
   printf("4\n");
   WrPortI(PFFR, NULL, 0x10);
   printf("2\n");
   WrPortI(PFDDR, NULL, 0xff);
   printf("8\n");

	WrPortI(TACR, &TACRShadow, 0x0C);	// port.timerA2-A7 Control Reg: PortA = mainClk,
										// Port E und F = port.timerA1
  	//printf("TACR set ...\n");
	WrPortI(TAT1R, NULL, 167);	// port.timerA1 auf 38,4 KHz
	//printf("TAT1R set ...\n");
	WrPortI(TAPR, NULL, 0x00);	// port.timerA Prescale Reg: mainClk = peripheralClk
  	//printf("TAPR set ...\n");
	WrPortI(TACSR, &TACSRShadow, 0x01);	// enable port.timerA and disable interrupts
  	//printf("TACSR set ...\n");
   report('s', "port.timer A set to 38.4 KHz (interrupts disabled)");

	// Seriel Port A (Debugging Port)
	WrPortI(TAT4R, NULL, 27);	// PortA:
	report('s', "Port A set to 115K2 ...");

	// Seriel Port E
	WrPortI(TAT2R, NULL, 3); // PortE:9K6Baud
	WrPortI(SECR, NULL, 0x02); // Interrupt priority 2
	WrPortI(SEER, NULL, 0x04); // AsynClk x8 Datarate
  	report('s', "Port E set to LOW-Speed 9K6 ...");

   // Seriel Port F
	WrPortI(TAT3R, NULL, 3); // PortF:9K6Baud
	WrPortI(SFCR, NULL, 0x02); // Interrupt priority 2
	WrPortI(SFER, NULL, 0x04); // AsynClk x8 Datarate
  	report('s', "Port F set to LOW-Speed 9K6 ...");


	report('s', "init sintab");
	printf("len %u\n", sizeof(sintab)/sizeof(float));
   for (l = 0; l < sizeof(sintab)/sizeof(float); l++) {
		sintab[l] = sin(3.1415926535*l/360.0);
   }

  	report('s', "Starting main loop ...");
	l = 0;
	for (;;) {

		costate {//some functions of the IP-system
		   tcp_tick(NULL);
         waitfor(DelayMs(80));
      }

      // Gibt einen Takt fuer Timeoutmessungen vor
      costate {
      	waitfor(DelayMs(STWAIT));
      	time++;//(our clock for timeouts)
      }

      // send telegrams to LMS with break of 1 ms
      costate {
      	while (porte.outbuf_in != porte.outbuf_out+1) { //Port E senden
				WrPortI(SEDR, NULL, porte.outbuf[++(porte.outbuf_out)]);
         	waitfor(DelayMs(1));
         }
			waitfor(DelayMs(20));
		}
      costate {
         while (portf.outbuf_in != portf.outbuf_out+1) { //Port F senden
				WrPortI(SFDR, NULL, portf.outbuf[++(portf.outbuf_out)]);
         	waitfor(DelayMs(1));
         }
			waitfor(DelayMs(20));
      }

      //check for buffer overrun
		costate {
			waitfor(DelayMs(20));
			if (((porte_in-(porte_out+1))%BUFFER_COUNT) > porte.maxfs)
	      	porte.maxfs = (porte_in-(porte_out+1))&BUFFER_MASK;
			if (((portf_in-(portf_out+1))%BUFFER_COUNT) > portf.maxfs)
	      	portf.maxfs = (portf_in-(portf_out+1))&BUFFER_MASK;
		}

      // work the packets that arrived from LMS
      costate {
         waitfordone do_packet(&porte);
         waitfordone do_packet(&portf);
      }

      // timeout for die LMS state automat
		costate {
      	waitfor(DelayMs(100));
			waitfordone do_timeout(&porte);
			waitfordone do_timeout(&portf);
      }

      // blinking of LEDs
      costate {
         waitfor(DelayMs(378));
         waitfordone blinken();
      }

      // check for UDP packets from host PC
      costate {
      	receive_packet();
         waitfor(DelayMs(400));
      }


/*
      costate {
      	waitfordone cpuburn();
      }
*/
/*
      costate {
			waitfor(DelayMs(5));
         waitfordone work_scan();
      }
*/

#if DEBUG
		costate {
 			waitfor(DelayMs(1000));
         fill_info(debug_buf, sizeof(debug_buf));
			printf("%s\n", debug_buf);
         porte.maxfs = 0;
         portf.maxfs = 0;
			l++;
 		}
#endif
	}
}






